#!/bin/bash

# Copyright 2016 SpeechLab at FEE CTU Prague
# Apache 2.0
# The script for training the ML AM

. ./cmd.sh

[ -f path.sh ] && . ./path.sh
set -e

# Set the path to the root speecon database directory
speechdat=/data/SPEECHDAT

# Choose language
language="CS HU RU PL SK"

# Set the working directory
dir_work=/workspace/fiala/recipes_ctu/kaldi/speechdat_ml

# stage of run.sh
stages=$1;
#      0 - Data & Lexicon & Language Preparation
#      1 - Merge data 
#      2 - Feature extraction
#      3 - AM training
#      4 - Create HCLG
#      5 - Decode
#      6 - DNN

tool="compute-mfcc"
#    - compute-mfcc # KALDI mfcc feature extraction tool
#    - ctucopy      # CtuCopy tool implementing speech enhancement and feature extraction algorithms.

#number of jobs for train
train_nj=15

for stage in $stages; do

    # Prepare data for every language individually
    if [ $stage == 0 ]; then
        for lang in $language; do
            echo "Data & Lexicon & Language Preparation"
            local/speechdat_data_prep.sh $speechdat $lang $dir_work
            utils/prepare_lang.sh data/$lang/local/dict "!SIL" \
                                  data/$lang/local/lang_tmp \
                                  data/$lang/lang || exit 1;
            bash local/create_LM.sh data/$lang/train/text \
                               data/$lang/local/lm || exit 1;
            local/speechdat_format_lms.sh --lm-srcdir "data/$lang/local/lm" \
                                          data/$lang/lang
        done 
    fi

    # Merge data languages
    if [ $stage == 1 ]; then
        echo "merging"
        local/speechdat_data_prep_ml.sh "$language" $dir_work
        utils/prepare_lang.sh data/ml/local/dict "!SIL"\
                              data/ml/local/lang_tmp \
                              data/ml/lang || exit 1;
#        local/create_LM.sh  --smoothing "-wbdiscount" \
#                            data/ml/train/text data/ml/local/lm || exit 1;
        bash local/create_LM.sh data/$lang/train/text \
                               data/$lang/local/lm || exit 1;
        local/speechdat_format_lms.sh --lm-srcdir "data/ml/local/lm"\
                                      data/ml/lang
    fi

    if [ $stage == 2 ]; then
        echo "Feature Extration & CMN stats for Training and Test set"
        mfccdir=/scratch/fiala/speechdat_multilingual/mfcc
        for x in train dev test; do
            steps/make_mfcc.sh --cmd "$train_cmd" \
                               --nj 10  data/ml/$x \
                               exp/mfcc/ml/$x \
                               $mfccdir || exit 1;
            steps/compute_cmvn_stats.sh data/ml/$x \
                                        exp/mfcc/ml/$x \
                                        $mfccdir/ || exit 1;
        done
    fi

    if [ $stage == 3 ]; then
        echo "MonoPhone Training & Decoding"
        steps/train_mono.sh --nj "$train_nj" \
                            --cmd "$train_cmd" \
                            data/ml/train \
                            data/ml/lang \
                            exp/ml/mono || exit 1;

        echo "tri1 : Deltas + Delta-Deltas Training & Decoding"
        steps/align_si.sh --boost-silence 1.25 \
                          --nj "$train_nj" \
                          --cmd "$train_cmd" \
                          data/ml/train \
                          data/ml/lang \
                          exp/ml/mono \
                          exp/ml/mono_ali || exit 1;

        steps/train_deltas.sh --cmd "$train_cmd" 2500 15000 
                              data/ml/train \
                              data/ml/lang \
                              exp/ml/mono_ali \
                              exp/ml/tri1 || exit 1;


        echo "tri2 : LDA + MLLT Training & Decoding"
        steps/align_si.sh --nj "$train_nj"\
                          --cmd "$train_cmd" \
                          data/ml/train \
                          data/ml/lang \
                          exp/ml/tri1 \
                          exp/ml/tri1_ali || exit 1;

        steps/train_lda_mllt.sh --cmd "$train_cmd" \
                                 --splice-opts "--left-context=3 --right-context=3" 2500 15000 \
                                 data/ml/train \
                                 data/ml/lang \
                                 exp/ml/tri1_ali \
                                 exp/ml/tri2 || exit 1;

        echo "tri3 : LDA + MLLT + SAT Training & Decoding"
        steps/align_si.sh --nj "$train_nj" \
                          --cmd "$train_cmd" \
                          --use-graphs true \
                          data/ml/train \
                          data/ml/lang \
                          exp/ml/tri2 \
                          exp/ml/tri2_ali || exit 1;

        steps/train_sat.sh --cmd "$train_cmd" \
                            4200 40000 \
                            data/ml/train \
                            data/ml/lang \
                            exp/ml/tri2_ali \
                            exp/ml/tri3 || exit 1;

        echo "SGMM2 Training & Decoding"
        steps/align_fmllr.sh --nj "$train_nj" \
                             --cmd "$train_cmd" \
                             data/ml/train \
                             data/ml/lang \
                             exp/ml/tri3 \
                             exp/ml/tri3_alia || exit 1;

        steps/train_ubm.sh --cmd "$train_cmd" \
                           400 \
                           data/ml/train \
                           data/ml/lang \
                           exp/ml/tri3_alia \
                           exp/ml/ubm4a || exit 1;

        steps/train_sgmm2.sh --cmd "$train_cmd" \
                             7000 9000 \
                             data/ml/train \
                             data/ml/lang \
                             exp/ml/tri3_alia \
                             exp/ml/ubm4a/final.ubm \
                             exp/ml/sgmm2_4a || exit 1;

        echo "MMI + SGMM2 Training & Decoding"
        steps/align_sgmm2.sh --nj "$train_nj" \
                             --cmd "$train_cmd" \
                             --transform-dir exp/ml/tri3_alia \
                             --use-graphs true \
                             --use-gselect true \
                             data/ml/train \
                             data/ml/lang \
                             exp/ml/sgmm2_4a \
                             exp/ml/sgmm2_4_alia || exit 1;

        steps/make_denlats_sgmm2.sh --nj "$train_nj" \
                                    --sub-split "$train_nj" \
                                    --acwt 0.2 \
                                    --lattice-beam 10.0 \
                                    --beam 18.0 \
                                    --cmd "$train_cmd" \
                                    --transform-dir exp/ml/tri3_alia \
                                    data/ml/train \
                                    data/ml/lang \
                                    exp/ml/sgmm2_4_alia \
                                    exp/ml/sgmm2_4_denlatsa || exit 1;

        steps/train_mmi_sgmm2.sh --acwt 0.2 \
                                 --cmd "$train_cmd" \
                                 --transform-dir exp/ml/tri3_alia \
                                 --boost 0.1 \
                                 --drop-frames true \
                                 data/ml/train \
                                 data/ml/lang \
                                 exp/ml/sgmm2_4_alia \
                                 exp/ml/sgmm2_4_denlatsa \
                                 exp/ml/sgmm2_4_mmi_b0.1a || exit 1;
    fi

    if [ $stage == 4 ]; then
        echo "Create HCLG.fst graphs"
        (
        $train_cmd JOB=1:1 exp/ml/mono/log/mkgraph.lm_tg.log \
            utils/mkgraph.sh --mono data/ml/lang_test_lm_tg exp/ml/mono exp/ml/mono/graph_lm_tg || exit 1;
        )&

        for x in tri1 tri2 tri3 sgmm2_4a; do
            (
            $train_cmd JOB=1:1 exp/ml/$x/log/mkgraph.lm_tg.log \
            utils/mkgraph.sh data/ml/lang_test_lm_tg exp/ml/$x exp/ml/$x/graph_lm_tg || exit 1;
            )&
        done
        wait
    fi

    if [ $stage == 5 ]; then
       echo "Decode"
       for eval in test dev; do
         (
          steps/decode.sh --nj "$decode_nj" --cmd "$decode_cmd" \
                          exp/ml/mono/graph_lm_tg \
                          data/ml/${eval} \
                          exp/ml/mono/decode_${eval}_lm_tg || exit 1;

          steps/decode.sh --nj "$decode_nj" --cmd "$decode_cmd" \
                          exp/ml/tri1/graph_lm_tg \
                          data/ml/${eval} \
                          exp/ml/tri1/decode_${eval}_lm_tg || exit 1;

          steps/decode.sh --nj "$decode_nj" --cmd "$decode_cmd" \
                          exp/ml/tri2/graph_lm_tg \
                          data/ml/${eval} \
                          exp/ml/tri2/decode_${eval}_lm_tg || exit 1;

         steps/decode_fmllr.sh --nj "$decode_nj" --cmd "$decode_cmd" \
                         exp/ml/tri3/graph_lm_tg \
                          data/ml/${eval} \
                          exp/ml/tri3/decode_${eval}_lm_tg || exit 1;

          steps/decode_sgmm2.sh --nj "$decode_nj" --cmd "$decode_cmd" \
                          --transform-dir exp/ml/tri3/decode_${eval}_lm_tg \
                          exp/ml/sgmm2_4a/graph_lm_tg \
                          data/ml/${eval} \
                          exp/ml/sgmm2_4a/decode_${eval}_lm_tg || exit 1;

          for iter in 1 2 3 4; do
            (
            steps/decode_sgmm2_rescore.sh --cmd "$decode_cmd" \
                          --iter $iter \
                          --transform-dir exp/ml/tri3/decode_${eval}_lm_tg \
                          data/ml/lang_test_lm_tg \
                          data/ml/${eval} \
                          exp/ml/sgmm2_4a/decode_${eval}_lm_tg \
                          exp/ml/sgmm2_4_mmi_b0.1a/decode_${eval}_lm_tg$iter || exit 1;
            )&
          done
        )&
       done
    fi

    if [ $stage == 6 ]; then
        echo "DNN"
        lang="ml"
        local/run_dnn.sh --lang $lang \
                          --gmmdir exp/$lang/tri3 \
                          --exp_name "-" \
                          --stage 2 \
                          --data_fmllr /scratch/fiala/speechdat_ml/DNN/ 

        echo "STAGE $stage finished"
    fi
done
